package BU::PSA::Restore::Client;

require 5.004;
require Exporter;

use Digest::MD5 qw(md5_hex);
use BU::PSA::Const;
use BU::PSA::Logger;
use BU::FileFunc;
use BU::PSA::CommonFunc;
use BU::PSA::DB;
use BU::XML::CommonFunc;
use BU::PSA::Objects::Client;
use BU::PSA::Objects::ClientLimits;
use BU::PSA::Restore::Domain;
use BU::PSA::Restore::IPMapping;
use BU::PSA::Restore::ActionLog;
use BU::PSA::Map;
use BU::PSA::Restore::CommonData qw(restoreReports restoreCustomButtons);

@ISA = qw(Exporter);
@EXPORT = qw(restoreClient);

sub restoreClient
{
	my ($clientNode, $map_fields) = @_;

	# get Client login
	my ($loginNode) = $clientNode->xql('./login');
	my $login = u2a($loginNode->getAttribute('name'));
	my $passwd = u2a($loginNode->getAttribute('password'));
	my $passwd_type = u2a($loginNode->getAttribute('PW_TYPE'));

	if ($PRETTYPRINT) {
		print_cl("Client $login");
	} else {
		$conf{'CLIENTS_PROCESSED_NUMBER'}++;
		printf_stdout("%s client '%s' [%s/%s]\n", $ProcessName, $login, $conf{'CLIENTS_PROCESSED_NUMBER'}, $conf{'CLIENTS_TOTAL_NUMBER'});
	}

	# get client pname
	my ($pnameField) = $clientNode->xql('./card/pname' . $fieldName);
	unless ($pnameField) {
		printf_stderr("Can not retrieve a personal name for client with '%s' login.".
			 			"The restoration for this client will be skipped.\n", $login);

		return '';
	}
	my $textNode = $pnameField->getFirstChild;
	unless ($textNode) {
		printf_stderr("An error occured during client restoration: unable retrive personal name for client %s\n", $login);
		return '';
	}
	my $pname = u2a($textNode->toString);

	my $client = new BU::PSA::Objects::Client($login, $pname, $passwd, $passwd_type);
	unless ($client) {
		printf_stderr("Unable to create %s object for client with login %s and personal name %s", 
			 'BU::PSA::Client', $login, $pname);
		return '';
	}	

	if ($client->error()) {
		warn $client->getErrorMsg();
		return '';
	}

	my $isExist = $client->Exists();
	my $clientId = $isExist ? $client->getDBId() : undef;

	if ($flags{RESTORE_MODE} eq 'CHECK') {
		if ($isExist) {
			unless ($flags{OVERWRITE}) {
				printf_stderr("Restoring for the '%s' client is impossible: client with such name already exists.\n",
							 	$login);

				markCollision($clientNode);
				return '';
			}
		}

		if (my ($node) = $clientNode->xql('./vhost_skeleton')) {
			my $cid = $node->getAttribute('src');
			$cid =~ s/^cid://;
			$node->setAttribute('destdir', $conf{'HTTPD_VHOSTS_D'} . "/.skel/$cid");
			hashXMLNodeByCID($node);
		}

		my $query = dbDoQuery('SELECT c.state FROM dom_level_usrs dlu, domains d, Cards c WHERE '
							. ' dlu.card_id=c.id AND d.id=dlu.dom_id AND d.name=' . dbQuote($login));

		if ($query && (my ($state) = $query->fetchrow_array())) {
			if ($state eq 'true') {
				printf_stderr("Restoring for the %s client is impossible: ".
							 "domain level user with such name already exists and his account is enabled.\n",
							 $login);
			} else {
				printf_stderr("WARNING: domain level user with such name already exists in the PSA system.\n");
			}

			markCollision($clientNode);
			return '';
		}

		{### check collision by login and pname
			my $query = dbDoQuery(sprintf('SELECT login, pname FROM clients WHERE (login=%s AND pname!=%s) or (login!=%s AND pname=%s)',
								  dbQuote($login), dbQuote($pname), dbQuote($login), dbQuote($pname)));

			if (my ($psa_login, $psa_pname) = $query->fetchrow_array()) {
				printf_stderr("Unable to restore client with login '%s' and personal name '%s': collision detected, client with login '%s' and personal name '%s' already exists.\n",
							  $login, $pname, $psa_login, $psa_pname);

				$flags{'CLIENT_COLLISION_DETECTED'} = 1;
				markCollision($clientNode);
				return '';
			}
		}

		$conf{'CLIENTS_USAGE'}++ unless $isExist;

		foreach my $logoNode ($clientNode->xql('./logo[@file != ""]')) {
			my $fake_file_name = md5_hex('logo_file_' . rand());
			$logoNode->setAttribute('destfile', $conf{LOGOS_REPOSITORY_D} . "/$fake_file_name");

			$logoNode->setAttribute('fake', $fake_file_name);
			hashXMLNodeByCID($logoNode);
		}	

		## check for installation of site app packages
		foreach my $sn ($clientNode->xql('./site_app_packages_pool/*')) {
			# check accordable SiteAppPackage installation
			my $app_pkg_id = $sn->getAttribute('id');
			my ($pkgNode) = $psaNode->xql('./server/site_app_packages/*[@id="'.$app_pkg_id.'"]');
			unless ($pkgNode) {
				print_stderr("Site application (with id='$app_pkg_id' in dump) will not be restored into client pool: unable to check package installation for site application.\n");
				&setRestoreFlags($baseNode, $parentNode, $domainNode, {'restore' => 'false'});
				next;
			}
			my $real_id = $pkgNode->getAttribute('real_id');
			### accordable package is not installed (real_id is defined in Config.pm)
			unless ($real_id) {
				printf_stderr("Site application '%s' will not be restored into client's pool because accordable site app package does not installed.\n", $pkgNode->getAttribute('name'));
				next;
			}
			$sn->setAttribute('real_id', $real_id);
		}
	} elsif ($flags{RESTORE_MODE} eq 'RESTORE') {
		if ($clientNode->getAttribute('collision')) {
			printf_stderr("'%s' client restoration is skipped.\n", $login);
			return '';
		}
	
		unless (defined($map_fields)) { $map_fields = \%card_fields; }

		foreach my $fieldName (values %$map_fields) {
			if ($fieldName eq 'pname') {
				$client->setField('pname', $pname);
			} elsif (my ($item) = $clientNode->xql('./card/' . $fieldName)) {
				my $textNode = $item->getFirstChild;
				$client->setField($fieldName, $textNode ? u2a($textNode->toString) : '');
			} else {
				$client->setField($fieldName, '');
			}
		}

		$client->setField('cr_date', $clientNode->getAttribute('create_date'));
		$client->setField('status', $clientNode->getAttribute('status'));

		if ($client->error()) {
			warn $client->getErrorMsg();
			return '';
		}

		$client->Store();

		if ($client->error()) {
			printf_stderr("Unable to save client info (login: '%s', personal name: '%s'): %s\n",
							$login, $pname, $client->getErrorMsg());
			return '';
		}

		$clientId = $client->getDBId();

		if (my ($logoNode) = $clientNode->xql('./logo')) {
			my $link = $logoNode->getAttribute('url');
			my $showMode = $logoNode->getAttribute('showMode');

			my $file = undef;
			my $fake = undef;
			if (-e $logoNode->getAttribute('destfile')) {
				$file = u2a($logoNode->getAttribute('file'));
				$fake = $logoNode->getAttribute('fake');
			}

			if ($file || $link) {
				unless ($file) {
					printf_stderr("Logo screen will be turned off in '%s' client preferences because ".
						 "the file %s was not restored.\n", $login, $file);
				}

				my $logo_id = $client->getField('logo_id');
				unless ($logo_id) {
					dbDoQuery("INSERT INTO Logos(name, fake, url, showMode) VALUES('$file', '$fake', '$link', '$showMode')");
					$logo_id = dbGetInsertId();
					$client->setField('logo_id', $logo_id);
				} else {
					dbDoQuery("UPDATE Logos set name='$file', fake='$fake', url='$link', showMode='$showMode' WHERE id=$logo_id");
				}
			}
		}

		my %ips_tmp = ();
		if (my ($node) = $clientNode->xql('./ip_pool')) {
			my $pool_id = $client->getField('pool_id');

			foreach my $ipNode ($node->xql('./ip_address')) {
				my $ip_address_id = $ipNode->getAttribute('id');
				my $ip_address_node = getIPAddressNodeById($ip_address_id);
				unless ($ip_address_node) {
					printf_stderr("Unable to restore ip address for client ip pool: unable to define ip address node by id\n");
					next;
				}

				### check "mapped" attribute for all ip's which was mapped
				### it is nessesary detecting ip addresses which must be restored for client ip pool 
				###### REASON:
				###### if ip address was mapped then ip pool collision was checked
				###### if ip address was not mapped then one don't have to restore in any case
				if ($ip_address_node->getAttribute('mapped') ne 'true') {
					next;
				}

				my $ip_address = $ip_address_node->getAttribute('ip_address');
				next if exists($ips_tmp{$ip_address});

				my $real_ip_address_id = $ip_address_node->getAttribute('real_id');
				unless ($real_ip_address_id) {
					printf_stderr("IP address '%s' restoration for client ip pool is skipped: ip address was not restored in system.\n"
									, $ip_address);
					next;
				}
				$ips_tmp{$ip_address} = 1;

				if ($pool_id) {
					dbDoQuery("REPLACE INTO Repository(rep_id, component_id) VALUES($pool_id, $real_ip_address_id)");
				} else {
					dbDoQuery("REPLACE INTO Repository(component_id) VALUES($real_ip_address_id)");
					$pool_id = dbGetInsertId();
					$client->setField('pool_id', $pool_id);
				}
			}

			if ($pool_id) {
				if (not $isExist) {
					restoreActionLog($node, $client->getDBId(), 'client_ip_pool');
				} else {
					logAction($client->getDBId(), 'client_ip_pool_update', undef);
				}
			}
		}

		my $sapp_pool_id = $client->getField('sapp_pool_id');
		my $sapp_pool_exists = $sapp_pool_id;
		foreach my $sn ($clientNode->xql('./site_app_packages_pool/*')) {
			my $real_id = $sn->getAttribute('real_id');
			my $app_pkg_id = $sn->getAttribute('id');
			next unless $real_id;
			if ($sapp_pool_exists) {
				my $res = dbDoQuery('SELECT COUNT(*) FROM Repository WHERE rep_id="' . $sapp_pool_id . '" AND component_id="' . $real_id . '"');
				unless ($res) {
					printf_stderr("Unable to check site application package existing in pool for client '%s'\n", $login);
				}
				my ($number) = $res->fetchrow_array();
				if ($number > 0) {
					my ($pkgNode) = $psaNode->xql('./server/site_app_packages/*[@id="'.$app_pkg_id.'"]');
					if ($pkgNode) {
						printf_stderr("Site application '%s' already exists in client's pool.\n", $pkgNode->getAttribute('name'));
					} else {
						printf_stderr("Instance of site application package with id='%d' already exists in client's pool.\n", $real_id);
					}
					next;
				}
			}
			if ($sapp_pool_id) {
				dbDoQuery('INSERT INTO Repository(rep_id, component_id) VALUES("' . $sapp_pool_id . '", "' . $real_id . '")');
			} else {
				dbDoQuery('INSERT INTO Repository(component_id) VALUES("' . $real_id . '")');
				$sapp_pool_id = dbGetInsertId();
			}
		}
		if (!$sapp_pool_exists && $sapp_pool_id) {
			$client->setField('sapp_pool_id', $sapp_pool_id);
			$client->Store();
		}

		if (my ($node) = $clientNode->xql('./vhost_skeleton')) {
			my $path = $node->getAttribute('destdir');
			if ( -d $path ) {
				my $old_path = $conf{'HTTPD_VHOSTS_D'} . "/.skel/$clientId";
				if ($isExist && $clientId && (-d $old_path)) {
					system("rm -rf $old_path");
				}
				print_stdout("Moving skeleton directory from '$path' to '$old_path' ... ");
				system("mv $path $old_path");
				print_stdout("done\n");
			}
		}

		#restore client Parameters
		if (my ($paramsNode) = $clientNode->xql('./params')) {
			my $paramsNodesList = $paramsNode->getAttributes();

			for (my $i=0; $i<$paramsNodesList->getLength; $i++) {
				my $paramNode = $paramsNodesList->item($i);
				my $name = $paramNode->getNodeName();
				my $value = u2a($paramNode->getNodeValue());
				dbDoQuery("REPLACE INTO cl_param(cl_id, param, val) VALUES($clientId, \"$name\", \"$value\")");
			}
		}

		#restore client Parameters
		if (my ($paramsNode) = $clientNode->xql('./ext_params')) {
			my $paramsNodesList = $paramsNode->getAttributes();

			my $param_id = $client->getField('param_id');
			for (my $i=0; $i<$paramsNodesList->getLength; $i++) {
				my $paramNode = $paramsNodesList->item($i);
				my $name = $paramNode->getNodeName();
				my $value = u2a($paramNode->getNodeValue());
				if ($param_id) {
					dbDoQuery("REPLACE INTO Parameters(id, parameter, value) VALUES($param_id, \"$name\", \"$value\")");
				} else {
					dbDoQuery("INSERT INTO Parameters(parameter, value) VALUES(\"$name\", \"$value\")");
					$param_id = dbGetInsertId();
					$client->setField('param_id', $param_id);
				}
			}
		}

		#restore reports for client
		restoreReports($clientNode, $clientId, $const{IS_CLIENT});

		unless ($isExist) {
			#restore client limits
			if (my ($limitsNode) = $clientNode->xql('./limits')) {
				my $limitsNodesList = $limitsNode->getAttributes();

				my $lim_id = $client->getField('limits_id');
				for (my $i=0; $i<$limitsNodesList->getLength; $i++) {
					my $limitNode = $limitsNodesList->item($i);
					my $name = $limitNode->getNodeName();
					my $value = u2a($limitNode->getNodeValue());

					if ($lim_id) {
						dbDoQuery("REPLACE INTO Limits(id, limit_name, value) VALUES($lim_id, \"$name\", \"$value\")");
					} else {
						dbDoQuery("INSERT INTO Limits(limit_name, value) VALUES(\"$name\", \"$value\")");
						$lim_id = dbGetInsertId();
						$client->setField('limits_id', $lim_id);
					}
				}

				restoreActionLog($limitsNode, $clientId, 'client_limits');
			}

			#restore client permissions
			if (my ($permsNode) = $clientNode->xql('./permissions')) {
				my $permsNodesList = $permsNode->getAttributes();

				my $perm_id = $client->getField('perm_id');
				for (my $i=0; $i<$permsNodesList->getLength; $i++) {
					my $permNode = $permsNodesList->item($i);
					my $name = $permNode->getNodeName();
					my $value = u2a($permNode->getNodeValue());
					if ($perm_id) {
						dbDoQuery("REPLACE INTO Permissions(id, permission, value) VALUES($perm_id, \"$name\", \"$value\")");
					} else {
						dbDoQuery("INSERT INTO Permissions(permission, value) VALUES(\"$name\", \"$value\")");
						$perm_id = dbGetInsertId();
						$client->setField('perm_id', $perm_id);
					}
				}

				restoreActionLog($limitsNode, $clientId, 'client_permissions');
			}
		}

		$client->Store();

		unless ($isExist) {
			restoreActionLog($clientNode, $client->getDBId(), 'client');
		} else {
			$client->logChanges($client->getDBId(), 'client_update');
		}

		foreach my $node ($clientNode->xql('./traffic_statistic/*')) {
			my $buff = XMLAttrs2SQL('REPLACE INTO ClientsTraffic(cl_id, %s) VALUES('.$clientId.', %s)', $node->getAttributes());
			dbDoQuery($buff);
		}

		my $rep_id = $client->getField('tmpl_id');
		foreach my $node ($clientNode->xql('./domain_templates/domain')) {
			my $tpl_id = undef;
			my $query = undef;

			if ($rep_id) {
				$query = dbDoQuery("SELECT tpl.id FROM Templates tpl, Repository r WHERE r.rep_id=$rep_id AND r.component_id=tpl.id AND name=" . dbQuote($node->getAttribute('name')));
				unless (($tpl_id) = $query->fetchrow_array()) {
					$tpl_id = undef;
				}
			}

			if ($tpl_id) {
				dbDoQuery("REPLACE INTO Templates(id, name) VALUES($tpl_id, ".dbQuote($node->getAttribute('name')).')');
			} else {
				dbDoQuery('INSERT INTO Templates(name) VALUES('.dbQuote($node->getAttribute('name')).')');
				$tpl_id = dbGetInsertId();
			}

			foreach my $sn ($node->xql('./*')) {
				my $field = $sn->getNodeName();
				my $value = undef;
				$sn->removeAttribute('id');
				if ($field eq 'log_rotation') {
					$field = 'logrotation_id';
					my $query = dbDoQuery('SELECT value FROM TmplData WHERE element="logrotation_id" AND tmpl_id=' . dbQuote($tpl_id));
					my ($logrot_id) = $query->fetchrow_array();
					if ($logrot_id) {
						my $buff = XMLAttrs2SQL('REPLACE INTO log_rotation(id, %s) VALUES('.$logrot_id.', %s)',
												$sn->getAttributes(), 1);
						dbDoQuery($buff);
						$value = $logrot_id;
					} else {
						my $buff = XMLAttrs2SQL('INSERT INTO log_rotation(%s) VALUES(%s)',
												$sn->getAttributes(), 1);
						dbDoQuery($buff);
						$value = dbGetInsertId();
					}
				} else {
					$value = $sn->getAttribute('value');
				}

				dbDoQuery('REPLACE INTO TmplData(tmpl_id, element, value) VALUES('.
							$tpl_id . ', "'. u2a($field) . '", "' . u2a($value) . '")');
					
			}

			if ($rep_id) {
				dbDoQuery("REPLACE INTO Repository(rep_id, component_id) VALUES($rep_id, $tpl_id)");
			} else {
				dbDoQuery("INSERT INTO Repository(component_id) VALUES($tpl_id)");
				$rep_id = dbGetInsertId();
				$client->setField('tmpl_id', $rep_id);
				$client->Store();
			}
		}
	}

	restoreCustomButtons($clientNode, $clientId, $const{'IS_CLIENT'}, $isExist);

	if ($conf{'RESTORING_BY_DOMAINS_LIST'}) {
		my $domainsListForRestore = $conf{'RESTORING_BY_DOMAINS_LIST'};
		foreach my $domainNode ($clientNode->xql('./sites/domain[@restore_by_list = "true"]')) {
			$conf{'DOMAIN_MAILNAMES_USAGE'} = 0;
			$conf{'DOMAIN_WEBUSERS_USAGE'} = 0;

			restoreDomain($clientId, $login, $domainNode);

			### all key limits set for each domain (it is not additive limit)
			if ($conf{'DOMAIN_MAILNAMES_USAGE'} > ($conf{'MAILNAMES_USAGE'} + 0)) {
				$conf{'MAILNAMES_USAGE'} = $conf{'DOMAIN_MAILNAMES_USAGE'};
			}
			if ($conf{'DOMAIN_WEBUSERS_USAGE'} > ($conf{'WEBUSERS_USAGE'} + 0)) {
				$conf{'WEBUSERS_USAGE'} = $conf{'DOMAIN_WEBUSERS_USAGE'};
			}
			delete $$domainsListForRestore{$domainNode->getAttribute('name')};
		}
	} else {
		foreach my $domainNode ($clientNode->xql('./sites/domain')) {
			$conf{'DOMAIN_MAILNAMES_USAGE'} = 0;
			$conf{'DOMAIN_WEBUSERS_USAGE'} = 0;

			restoreDomain($clientId, $login, $domainNode);

			### all key limits set for each domain (it is not additive limit)
			if ($conf{'DOMAIN_MAILNAMES_USAGE'} > ($conf{'MAILNAMES_USAGE'} + 0)) {
				$conf{'MAILNAMES_USAGE'} = $conf{'DOMAIN_MAILNAMES_USAGE'};
			}
			if ($conf{'DOMAIN_WEBUSERS_USAGE'} > ($conf{'WEBUSERS_USAGE'} + 0)) {
				$conf{'WEBUSERS_USAGE'} = $conf{'DOMAIN_WEBUSERS_USAGE'};
			}
		}
	}

	if ($flags{RESTORE_MODE} eq 'CHECK') {
		if ($client->Exists()) {
			unless (checkClientLimits($clientNode, $client)) {
				return '';
			}
		}
	}

	return 1;
}


sub checkClientLimits($ $)
{
	my ($clientNode, $client) = @_;

	my $clLimits = new BU::PSA::Objects::ClientLimits($client->getDBId(), $client->getField('limits_id'));
	if ($clLimits->error()) {
		print_stderr($clLimits->getErrorMsg() . "\n");
		return '';
	}

	$clLimits->fetchList();
	if ($clLimits->error()) {
		print_stderr($clLimits->getErrorMsg() . "\n");
		return '';
	}

	my %notAdditiveLimit = (map {$_ => $_} 
							'expiration', 'mbox_quota');

	my $client_expiration = $clLimits->getValue('expiration');

	my %summaryLimits = ();
	my $max_dom = 0;
	my $restore_by_list = ($clientNode->getAttribute('restore_by_list') ne 'true') ? '' : ' $and$ @restore_by_list="true"';
	foreach my $domNode ($clientNode->xql('./sites/domain[($not$ @exists) $and$ ($not$ @collision) '.$restore_by_list.']')) {
		foreach my $node ($domNode->xql('./limits')) {
			my $limitsNodesList = $node->getAttributes();

			for (my $i=0; $i<$limitsNodesList->getLength; $i++) {
				my $limitNode = $limitsNodesList->item($i);
				my $name = $limitNode->getNodeName();

				my $value = u2a($limitNode->getNodeValue()) + 0;

				if (($name eq 'expiration') && ($client_expiration > 0) && (($value > $client_expiration) || ($value < 0))) {
					$value = $client_expiration;
					$limitNode->setNodeValue(a2u($client_expiration));
					printf_stderr("Domain '%s' expiration date is later than client expiration date: set to client expiration date.\n",
								  $domNode->getAttribute('name'));
				}
				### ($value < 0) -> limit is unlimited
				### for not additive limits select max value
				if (($value < 0) || !exists($summaryLimits{$name})) {
					$summaryLimits{$name} = $value;
				} elsif (exists($notAdditiveLimit{$name})) {
					if ($value > ($summaryLimits{$name} + 0)) {
						$summaryLimits{$name} = $value;
					}
				} elsif ($summaryLimits{$name} >= 0) {
					$summaryLimits{$name} = $summaryLimits{$name} + $value;
				}
			}
		}
		$max_dom += 1;
	}
	$summaryLimits{'max_dom'} = $max_dom;

	my $limits_is_exceeded = '';

	foreach my $name (keys %summaryLimits) {
		my $dumpLimitValue = $summaryLimits{$name};

		printf "$name(%s)(%s)(%s)\n", $clLimits->getValue($name), $dumpLimitValue, $clLimits->getSummaryDomainsLimit($name) if ($debug > 3);

		my $clSummaryLimit = $clLimits->getSummaryDomainsLimit($name);

		my $tmp;
		if (($dumpLimitValue < 0) || ($clSummaryLimit < 0)) {
			$tmp = -1;
		} elsif (exists($notAdditiveLimit{$name})) {
    		### for not additive limits we need to check only max value in dump
			$tmp = $dumpLimitValue;
		} else {
			$tmp = $dumpLimitValue + $clSummaryLimit;
		}

		if (($clLimits->getValue($name) >= 0) && (($clLimits->getValue($name) < $tmp) || ($tmp < 0))) {
			printf_stderr("Error: client '%s' limit '%s' is %s, but summary domains limit required for restore: %s (in addition to %s already exists in Plesk)\n",
						  $client->getField('login'),
						  $name,
						  ($clLimits->getValue($name) < 0) ? 'unlimited' : $clLimits->getValue($name),
						  ($dumpLimitValue < 0) ? 'unlimited' : $dumpLimitValue,
						  ($clSummaryLimit < 0) ? 'unlimited' : $clSummaryLimit);

			$limits_is_exceeded = 1;
		}
	}

	if ($limits_is_exceeded) {
	    printf_stderr("Client '%s' will not be restore because summary domains limits will exceed.\n", 
					  $client->getField('login'));
	    markCollision($clientNode);
	    return '';
	}

	return 1;
}


return 1;

