package BU::PSA::Restore::Database;

require 5.004;
require Exporter;

use BU::PSA::Const;
use BU::PSA::Logger;
use BU::FileFunc;
use BU::PSA::CommonFunc;
use BU::PSA::DB;
use BU::PSA::Objects::Database;
use BU::PSA::Objects::DatabaseUser;
use BU::XML::CommonFunc;

@ISA = qw(Exporter);
@EXPORT = qw(restoreDatabase restoreDatabaseUser);

sub restoreDatabase($ $ $)
{
	my ($domainId, $domainName, $dbNode) = @_;

	my $dbName = u2a($dbNode->getAttribute('name'));
	my $dbType = $dbNode->getAttribute('type');

	printf_stdout("%s %s database '%s'\n", $ProcessName, $dbType, $dbName);

	# if $domainId is undefined than domain is not exist and database is not exist the same!
	my $isExist = '';
	my $db = undef;
	if ($domainId) {
		$db = new BU::PSA::Objects::Database($domainId, $dbName, $dbType);
		unless ($db) {
			print_stderr("Unable to create \"BU::PSA::Database\" object.\n");
			return '';
		}

		if ($db->error()) {
			printf_stderr("Error in database object for the database %s: %s\n", $dbName, $db->getErrorMsg());
			return '';
		}
		$isExist = $db->Exists();
	}

	if ($flags{RESTORE_MODE} eq 'CHECK') {
        hashXMLNodeByCID($dbNode);

		if ($isExist) {
			$dbNode->setAttribute('exists', 'true');
		}

		if (($dbType eq 'postgresql') && !$conf{PG_IS_RUNNING}) {
			printf_stderr("Database %s will not be restored: postgres service is not running now\n", $dbName);
			markCollision($dbNode);
			return 1;
		}

		{  ### check for database collision
			my $query = dbDoQuery(sprintf('SELECT d.name FROM data_bases db, domains d WHERE d.id=db.dom_id AND db.type=%s AND lower(db.name)=lower(%s)',
											, dbQuote($dbType), dbQuote($dbName)));

			my ($dbDomainName) = $query->fetchrow_array if $query;

			unless ($query || $dbDomainName) {
				print_stderr("Unable to define database depending on domain.\n");
			} elsif ($dbDomainName && ($dbDomainName ne $domainName)) {
				printf_stderr("Database %s restoration is impossible.\n", $dbName);
				printf_stderr("Restoring for the %s database is impossible: ".
							 "database with such name already exists but it is related to the %s domain\n",
							 $dbName, $dbDomainName);
				markCollision($dbNode);
				return '';
			}
		}
	} elsif ($flags{RESTORE_MODE} eq 'RESTORE') {

		if ($dbNode->getAttribute('collision') || ($dbNode->getAttribute('restore') eq 'false')) {
			print_stderr("Database $dbName restoration is skipped.\n");
			return '';
		}

		$db->Store();

		if ($db->error()) {
			printf_stderr("ERROR: Unable to create database %s: %s\n", $dbName, $db->getErrorMsg());
			return '';
		}

		# set real_id attribute: It required for site application restoration
		$dbNode->setAttribute('real_id', $db->getDBId());

		if ($dbType eq 'mysql') {
			my $is_mysql4 = (ver2int($const{MYSQL_VERSION}) >= ver2int('4.0.0'));
			foreach my $accessRightsNode ($dbNode->xql('./access_rights')) {
				### BUGFIX #38109 mysql4 privileges supporting
				my @EXTRA_MYSQL_DB_PRIVILEGES = ('Create_tmp_table_priv', 'Lock_tables_priv');
				if ($is_mysql4) {
					### if it is mysql4 and extra privileges is not defined then define it
					foreach my $extra_priv_name (@EXTRA_MYSQL_DB_PRIVILEGES) {
						$accessRightsNode->setAttribute($extra_priv_name, 'Y') unless $accessRightsNode->getAttribute($extra_priv_name);
					}
				} else {
					### if it is mysql3 and extra priv is defined then remove it
					foreach my $extra_priv_name (@EXTRA_MYSQL_DB_PRIVILEGES) {
						$accessRightsNode->removeAttribute($extra_priv_name) if $accessRightsNode->getAttribute($extra_priv_name);
					}
				}
				my $nodesList = $accessRightsNode->getAttributes();
				my $buff = XMLAttrs2SQL('REPLACE INTO mysql.db(Db, %s) VALUES('.dbCorrectDBName($dbName).', %s)',
										$nodesList, 1);
				dbDoQuery($buff);
			}
		}
	}

	foreach my $dbUserNode ($dbNode->xql('./user[@type="db_users"]')) {
		restoreDatabaseUser($db ? $db->getDBId() : undef, $dbName, $dbType, $dbUserNode);
	}

	return 1;
}


sub restoreDatabaseUser($ $ $ $)
{
	my ($dbId, $dbName, $dbType, $dbUserNode) = @_;

	my ($dbLoginNode) = $dbUserNode->xql('./login');
	
	unless ($dbLoginNode) {
		print_stderr("Unable to restore database user: can not retrive user name.\n");
		return '';
	}

	my $dbUserName = u2a($dbLoginNode->getAttribute('name'));

	printf_stdout("%s %s database user '%s'\n", $ProcessName, $dbType, $dbUserName);

	# if $dbId is not defined than database is not exist and db user is not exist in PSA db table!
	my $isExist = '';
	my $dbUser = undef;
	if ($dbId) {
		$dbUser = new BU::PSA::Objects::DatabaseUser(
										$dbId,
										$dbName, $dbType,
										$dbUserName,
										$dbLoginNode->getAttribute('PW_TYPE'),
										u2a($dbLoginNode->getAttribute('password'))
									);
		unless ($dbUser) {
			print_stderr("Unable to create \"BU::PSA::DatabaseUser\" object.\n");
			return '';
		}

		if ($dbUser->error()) {
			print_stderr("ERROR: Unable to create database user %s: %s\n", $dbuserName, $dbUser->getErrorMsg());
			return '';
		}
		$isExist = $dbUser->Exists();
	}

	if ($flags{RESTORE_MODE} eq 'CHECK') {

		if ($isExist) {
			$dbUserNode->setAttribute('exists', 'true');
		}

		{  ### check for database users collision
			my $query = dbDoQuery('SELECT db.name FROM data_bases db, db_users usr WHERE db.id=usr.db_id'
								. ' and lower(usr.login)=lower(' . dbQuote($dbUserName) . ')');
			
			my ($userDBName) = $query->fetchrow_array if $query;

			unless ($query || $userDBName) {
				print_stderr("Unable to check whether a database user related to the database %s ".
					 "exists in the PSA system.\n", $dbName);
			} elsif ($userDBName && ($userDBName ne $dbName)) {
				printf_stderr("'%s' database user restoration is impossible: such user already ".
					 "exists in PSA but he/she is related to the database %s\n",
					 $dbUserName, $userDBName);

				markCollision($dbUserNode);
				return '';
			}
		}

	} elsif ($flags{RESTORE_MODE} eq 'RESTORE') {

		if ($dbUserNode->getAttribute('collision') eq 'true' || ($dbUserNode->getAttribute('restore') eq 'false')) {
			print_stderr("'$dbUserName' database user restoration is skipped.\n");
			return '';
		}

		if ($dbUserNode->getAttribute('not_psa_user') ne 'true') {
			$dbUser->Store();

			if ($dbUser->error()) {
				printf_stderr("ERROR: Unable to create database user %s: %s\n", $dbuserName, $dbUser->getErrorMsg());
				return '';
			}

			# set real_id attribute: It required for site application restoration
			$dbLoginNode->setAttribute('real_id', $dbUser->getDBId());
		}

		if ($dbType eq 'mysql') {
			my $is_mysql4 = (ver2int($const{MYSQL_VERSION}) >= ver2int('4.0.0'));
			foreach my $accessRightsNode ($dbUserNode->xql('./access_rights')) {
				### BUGFIX #38109 mysql4 privileges supporting
				my @EXTRA_MYSQL_USER_PRIVILEGES = ('Show_db_priv', 'Super_priv', 'Create_tmp_table_priv', 'Lock_tables_priv', 'Execute_priv', 'Repl_slave_priv', 'Repl_client_priv');
				if ($is_mysql4) {
					### if it is mysql4 and extra privileges is not defined then define it
					foreach my $extra_priv_name (@EXTRA_MYSQL_USER_PRIVILEGES) {
						$accessRightsNode->setAttribute($extra_priv_name, 'N') unless $accessRightsNode->getAttribute($extra_priv_name);
					}
				} else {
					### if it is mysql3 and extra priv is defined then remove it
					foreach my $extra_priv_name (@EXTRA_MYSQL_USER_PRIVILEGES) {	
						$accessRightsNode->removeAttribute($extra_priv_name) if $accessRightsNode->getAttribute($extra_priv_name);
					}
				}
				my $nodesList = $accessRightsNode->getAttributes();
				my $buff = XMLAttrs2SQL('REPLACE INTO mysql.user(User, %s) VALUES('.dbQuote($dbUserName).', %s)',
										$nodesList, 1);

				dbDoQuery($buff);
			}
		}
	}

	return 1;
}



return 1;

