package BU::PSA::Restore::Hosting;

require 5.004;
require Exporter;

use BU::PSA::Const;
use BU::PSA::Logger;
use BU::FileFunc;
use BU::PSA::CommonFunc;
use BU::PSA::DB;
use BU::PSA::Objects::Hosting;
use BU::PSA::Restore::IPMapping;
use BU::PSA::Restore::PDir;
use BU::PSA::Restore::WebUser;
use BU::PSA::Restore::AnonFTP;
use BU::PSA::Restore::SubDomain;
use BU::PSA::Restore::SiteApp;
use BU::PSA::Restore::ActionLog;
use BU::XML::CommonFunc;

@ISA = qw(Exporter);
@EXPORT = qw(restorePhysicalHosting);

sub restorePhysicalHosting($ $ $ $ $; $ $)
{
	my ($domainId, $domainName, $domainDisplayName, $domain, $hostingNode, $single_domain_mode, $single_domain_ip_id) = @_;
	my $ftpNode = undef;

	$single_domain_mode = '' unless(defined($single_domain_mode));

	printf_stdout("%s hosting for domain '%s'\n", $ProcessName, $domainName);


	unless (($ftpNode) = $hostingNode->xql('./login')) {
		print_stderr("An error occured during physical hosting restoration: ".
			 "unable to get xml node for the ftp login.\n");
		return '';
	}

	my $ftpLogin = u2a($ftpNode->getAttribute('name'));
	my $ftpHomeDir = mapSysUserHomeDir(u2a($ftpNode->getAttribute('home')));
	$ftpNode->setAttribute('home', a2u($ftpHomeDir)) if $ftpHomeDir;

	if ($flags{RESTORE_MODE} eq 'CHECK') {
		$hostingNode->setAttribute('destdir', $conf{HTTPD_VHOSTS_D} . "/$domainName");
        hashXMLNodeByCID($hostingNode);
		if (my ($systemDataNode) = $hostingNode->xql('./system_data')) {
			$systemDataNode->setAttribute('destdir', $conf{HTTPD_VHOSTS_D} . "/$domainName");
        	hashXMLNodeByCID($systemDataNode);
		}

		if (my ($crontabNode) = $hostingNode->xql('./crontab')) {
        	hashXMLNodeByCID($crontabNode);
		}

		if (my ($certNode) = $hostingNode->xql('./rootchain_certificate')) {
			$certNode->setAttribute('destfile', $conf{HTTPD_VHOSTS_D} . "/$domainName/cert/rootchain.pem");
        	hashXMLNodeByCID($certNode);
		}

		if (my ($fpNode) = $hostingNode->xql('./services/fp')) {
			$fpRequired = 1;
		}

		{ ### check for hosting on other domain with such ftp user
			my $query = dbDoQuery('SELECT d.name FROM hosting h, domains d, sys_users su WHERE h.dom_id=d.id'
								. ' AND h.sys_user_id=su.id AND su.login=' . dbQuote($ftpLogin));

			my ($hostingDomainName) = $query->fetchrow_array if $query;

			unless ($query) {
				printf_stderr("Unable to check whether a %s domain-related FTP user for hosting exists ".
					 "in the PSA system\n", $domainName);
			} elsif ($hostingDomainName && ($domainName ne $hostingDomainName)) {
				printf_stderr("Hosting restoration for the '%s' domain is impossible: ".
					 "the hosting ftp user '%s' already exists but he/she is related to domain '%s'.\n",
					 $domainName, $ftpLogin, $hostingDomainName);
				markCollision($hostingNode);
				return '';
			}
		}

		{ ### compare hosting ftp user login with login in dump
			my $query = dbDoQuery('SELECT su.login FROM hosting h, domains d, sys_users su WHERE h.dom_id=d.id'
								. ' AND h.sys_user_id=su.id AND d.name=' . dbQuote($domainName));

			my ($hostingFtpLogin) = $query->fetchrow_array if $query;

			unless ($query) {
				printf_stderr("Unable to check whether a %s domain-related FTP user for hosting exists ".
					 "in the PSA system.\n", $domainName);
			} elsif ($hostingFtpLogin && (lc($ftpLogin) ne lc($hostingFtpLogin))) {
				printf_stderr("Hosting restoration for the '%s' domain is impossible: ".
					 "the hosting for this domain already exists but ftp login is differ from dump \n",
					 $domainName, $ftpLogin, $hostingDomainName);
				markCollision($hostingNode);
				return '';
			}

			if ($hostingFtpLogin) {
				################################################
				### USE FTP LOGIN NAME AS IN DATABASE
				$ftpLogin = $hostingFtpLogin;
				$ftpNode->setAttribute('name', a2u($ftpLogin));
				################################################
			}
		}

		if (util_exec('usermng', 'exist', $ftpLogin)) {
			printf_stdout("'%s' user was found in the system\n", $ftpLogin);

			my @pw_info = getpwnam($ftpLogin);

			if (lc($pw_info[7]) ne lc($ftpHomeDir)) {
				printf_stderr("Hosting restoration for the domain '%s' is impossible: "
							. "such user already exists in the system and he/she in not "
							. "a valid PSA ftp user for hosting.\n", $domainName);

				markCollision($hostingNode);
				return '';
			}
			$ftpHomeDir = $pw_info[7];
			$ftpNode->setAttribute('home', a2u($ftpHomeDir)) if $ftpHomeDir;
		}
	} elsif ($flags{RESTORE_MODE} eq 'RESTORE') {

		if ($hostingNode->getAttribute('collision') eq 'true') {
			printf_stderr("Hosting restoration for '%s' domain is skipped.\n", $domainName);
			return '';
		}

		my $ip_address = undef;
		my $ip_address_id = undef;
		if ($single_domain_mode) { # use exist ip address if it is single domain mode
			print_stdout("RESTORE SINGLE_DOMAIN_MODE (ip id:$single_domain_ip_id).\n") if $debug > 2;
			unless ($single_domain_ip_id) {
				printf_stderr("Hosting restoration for '%s' domain is skipped: can't determine ip address for hosting.\n", $domainName);
				return '';
			}
			my $ipAddrNode = getIPAddressNodeById($single_domain_ip_id);
			$ip_address = $ipAddrNode->getAttribute('ip_address');
			$ip_address_id = $ipAddrNode->getAttribute('real_id');
		} else {
			if (my($ip_address_node) = $hostingNode->xql('./ip_address')) {
				my $ipAddrNode = getIPAddressNodeById($ip_address_node->getAttribute('id'));
				$ip_address = $ipAddrNode->getAttribute('ip_address');
				$ip_address_id = $ipAddrNode->getAttribute('real_id');
			}
		}

		my $hosting = new BU::PSA::Objects::Hosting(
									$domainId,
									$domainName,
									$domainDisplayName,
									$ip_address,
									$ftpLogin,
									u2a($ftpNode->getAttribute('password')),
									$ftpNode->getAttribute('PW_TYPE'),
									$ftpHomeDir,
									u2a($ftpNode->getAttribute('shell')),
									$ftpNode->getAttribute('quota')
								);

		unless ($hosting) {
			printf_stderr("Unable to create %s object for domain '%s'.\n", "BU::PSA::Hosting\n", $domainName);
			return '';
		}

		if ($hosting->error()) {
			printf_stderr($hosting->getErrorMsg());
			return '';
		}

		my $isExist = $hosting->Exists();

		$hosting->setField('ip_address_id', $ip_address_id);
		$hosting->setField('at_domains', $hostingNode->getAttribute('at_domains'));

		if (my ($certNode) = $hostingNode->xql('./rootchain_certificate')) {
			my $status = $certNode->getAttribute('status');

			if ($status eq 'true') {
				if (-e $certNode->getAttribute('destfile')) {
					system("chown root:0 ".$certNode->getAttribute('destfile'));
					chmod 0400, $certNode->getAttribute('destfile');
				} else {
					printf_stderr("ERROR: rootchain certificate for domain %s is disabled: "
						."file with certificate is not exists.\n", $domainName);
					$status = 'false';
				}
			}
			$hosting->setField('rootchain', $status);
		}

		if (my ($trafficNode) = $hostingNode->xql('./traffic')) {
			if (my $real_traffic = $trafficNode->getAttribute('real')) {
				$hosting->setField('real_traffic', $real_traffic);
			}
#			my $traffic_limit = $trafficNode->getAttribute('limit');
#
#			if ($traffic_limit && ($traffic_limit != 'unlimited')) {
#				my $lim_id = $domain->getParam('lim_id');
#				unless ($lim_id) {
#					dbDoQuery('INSERT INTO Limits(limit_name, value) VALUES("max_traffic", ' . $traffic_limit . ')');
#					$domain->setParam('lim_id', dbGetInsertId());
#				} else {
#					dbDoQuery('UPDATE Limits SET value=' . $traffic_limit . ' WHERE limit_name="max_traffic" AND id=' . $lim_id);
#				}
#			}
		}

		my %fpUsersList = ();
		my $fpAuthOn = undef;
		if (my ($servicesNode) = $hostingNode->xql('./services')) {
			my $serviceNodesList = $servicesNode->getAttributes();

			for (my $i=0; $i<$serviceNodesList->getLength; $i++) {
				my $serviceNode = $serviceNodesList->item($i);
				$hosting->setField($serviceNode->getNodeName(), u2a($serviceNode->getNodeValue()));
			}

			unless ($conf{'COLDFUSION_IS_INSTALLED'}) {
				$hosting->setField('coldfusion', 'false');
			}

			my ($fpNode) = $servicesNode->xql('./fp');
			if ($fpNode) {
				if ($conf{'FRONTPAGE_IS_INSTALLED'}) {
					$hosting->setField('fp', $fpNode->getAttribute('not_ssl'));
					$hosting->setField('fp_ssl', $fpNode->getAttribute('ssl'));
				} else {
					$hosting->setField('fp', 'false');
					$hosting->setField('fp_ssl', 'false');
					if (($fpNode->getAttribute('not_ssl') eq 'true') || ($fpNode->getAttribute('ssl') eq 'true')) {
						printf_stderr("FrontPage service is turned off because FrontPage is not installed.\n");
					}
				}
				$hosting->setField('fp_enable', $fpNode->getAttribute('fp_auth'));
				$fpAuthOn = $conf{'FRONTPAGE_IS_INSTALLED'};

				my $fpLogin, $fpPassword, $fpPasswordType;
				foreach my $fpLoginNode ($fpNode->xql('./user/login')) {
					$fpLogin =  u2a($fpLoginNode->getAttribute('name'));
					$fpPassword = u2a($fpLoginNode->getAttribute('password'));
					$fpPasswordType = $fpLoginNode->getAttribute('PW_TYPE');
					my $isAdmin = ($fpLoginNode->getAttribute('ADMIN') eq 'true');
					$hosting->setField('fp_adm', $fpLogin) if $isAdmin;
					if ($fpPasswordType eq 'plain') {
						$hosting->setField('fp_pass', $fpPassword) if $isAdmin;
						$fpUsersList{$fpLogin} = crypt($fpPassword, genSaltForCrypt());
					} else {
						$hosting->setField('fp_pass', $fpLogin) if $isAdmin;
						$fpUsersList{$fpLogin} = $fpPassword;
					}
				}
				unless ($hosting->getField('fp_adm')) {
					printf_stderr("FrontPage restoration error: frontpage admin not found in dump. Use '$fpLogin' user as admin.\n");
					$hosting->setField('fp_adm', $fpLogin);
					$hosting->setField('fp_pass', ($fpPasswordType eq 'plain') ? $fpPassword : $fpLogin);
				}
			}
		}

		if ($hosting->error()) {
			print_stderr('Hosting restoration error occurs: ' . $hosting->getErrorMsg());
			return '';
		}

		$result = $hosting->Store();

		if ($hosting->error()) {
			printf_stderr("Error occured during domain '%s' Hosting restoration: %s\n", $domainName, $hosting->getErrorMsg());
			return '' unless $result;
		}

		my %dirs = {
				'httpdocs' => ($hosting->getField('fp') eq 'true'),
				'httpsdocs' => ($hosting->getField('fp_ssl') eq 'true')
		};

		foreach my $http_dir (keys %dirs) {
			next unless $dirs{$http_dir};

			my $fp_path = '/usr/local/frontpage';
			if ( ! -d $fp_path ) {
				$fp_path = $conf{PRODUCT_ROOT_D} . '/frontpage';
			}

			system("chown $ftpLogin:".$const{'APACHE_GROUP'}." $fp_path/$domainName:*.cnf >/dev/null 2>&1");

			my $search_path = $conf{HTTPD_VHOSTS_D} . "/$domainName/$http_dir";
			my $domainName_expr = $domainName;
			$domainName_expr =~ s/\./\\\./g;
			my $new_path_expr = $conf{HTTPD_VHOSTS_D};

			$expression = '^(\s*(Auth[A-Za-z]*File|PasswordDir\s*:)[^/]*).*(\/' . $domainName_expr . '\/.*)$';

			if (open(fp_fd, 'find '.$search_path.' -type f | grep -e "\(\.htaccess\|access\.cnf\)" 2>/dev/null|')) {
				while (<fp_fd>) {
					my $fname = $_;
					chomp($fname);

					unless (open (fp_conf_fd, $fname)) {
						print_stderr "Unable to open frontpage config file '$_': $!\n";
						next;
					}
					unless (open (fp_conf_fd_tmp, ">$fname.tmp")) {
						close fp_conf_fd;
						print_stderr "Unable to open tmp file '$fname.tmp': $!\n";
						next;
					}
					while (<fp_conf_fd>) {
						s/$expression/\1$new_path_expr\3/g;
						print fp_conf_fd_tmp "$_";
					}
					if (( close fp_conf_fd ) && ( close fp_conf_fd_tmp )) {
						system("cp -f '$fname.tmp' '$fname' 2>/dev/null");
						unlink("$fname.tmp") if (-e "$fname.tmp");
						system("chown $ftpLogin:" . $const{'APACHE_GROUP'} . " \"$fname\"") if (-e $fname);
					}
				}
				close fp_fd;
			} else {
				print_stderr "Unable to fetch frontpage config files list (.htacess, access.conf): $!\n";
			}
		}

		if ($fpAuthOn) {
			foreach my $dir_name (keys %dirs) {
				next unless $dirs{$dir_name};

				if (open fp_pwd_file, '+>' . $conf{HTTPD_VHOSTS_D} . "/$domainName/httpdocs/_vti_pvt/service.pwd") {

					print fp_pwd_file "# -FrontPage-\n";

					while (my ($fpLogin, $fpPassword) = each %fpUsersList) {
						print fp_pwd_file "$fpLogin:$fpPassword\n";
					}

					close fp_pwd_file;
				} else {
					printf_stderr("Unable to add FrontPage users, unable to create Frontpage password file: %s\n", $!);
				}
			}
		}

		if ($hosting->error()) {
			printf_stderr($hosting->getErrorMsg()."\n");
			return 1;
		}

		### restore action log
		unless ($isExist) {
			restoreActionLog($hostingNode, $domainId, 'phys_hosting');
		} else {
			$hosting->logChanges($domainId, 'phys_hosting_update');
		}

		# restore Protected directories
		foreach my $pDirNode ($hostingNode->xql('./protected_dir')) {
			restorePDir($domainId, $domainName, $ftpLogin, $pDirNode);
		}

		#restore anonymous FTP service
		if (my ($anonFtpNode) = $hostingNode->xql('./anon_ftp')) {
			restoreAnonFTP($domainId, $domainName, $anonFtpNode, $ip_address_id);
		}
	}

	# restore WebUsers
	foreach my $webUserNode ($hostingNode->xql('./user[@type = "web_users"]')) {
		restoreWebUser($domainId, $domainName, $webUserNode);
	}

	# restore SiteApps
	my $domainNode = $hostingNode->getParentNode();
	if ($domainNode->getNodeName() ne 'domain') {
		print_stderr("FATAL error: hosting parent node is not 'domain', it is '".$domainNode->getNodeName()."'\n");
		die "\n";
	}
	foreach my $siteAppNode ($hostingNode->xql('./site_apps/*')) {
		if (($flags{RESTORE_MODE} eq 'CHECK') && ($hostingNode->getAttribute('content_restore') eq 'false')) {
			markCollision($siteAppNode);
		} else {
			restoreSiteApp($siteAppNode, $domainNode, $domainId, $domainName);
		}
	}
	if (($flags{RESTORE_MODE} eq 'CHECK') && ($hostingNode->getAttribute('content_restore') eq 'false')) {
		print_stderr("Hosting content and all site applications for doman '$domainName' will not be restored because site application collision detected\n");
		if (my ($systemDataNode) = $hostingNode->xql('./system_data')) {
			$systemDataNode->setAttribute('content_restore', 'false');
		}
	}

	if (($flags{RESTORE_MODE} eq 'RESTORE') && $ftpNode) {
		my $uid = $ftpNode->getAttribute('uid'); 
		if (($uid ne '') && ($uid+0 != getpwnam($ftpLogin)+0)) {
			system("find $ftpHomeDir \\( -user $uid -or -nouser \\) -and \\( -type f -or -type d \\) -exec chown $ftpLogin {} \\; > /dev/null 2>&1");
		} else {
			system("find $ftpHomeDir -nouser -and \\( -type f -or -type d \\) -exec chown $ftpLogin {} \\; > /dev/null 2>&1");
		}
	}

	# restore Subdomains (it should be executed after chowning)
	foreach my $baseNode ($hostingNode->xql('./subdomains/*')) {
		restoreSubDomain($baseNode, $domainId, $domainName, $ftpNode);
	}

	return 1;
}


return 1;

